import re

class Assembler:
    class AssemblyException(Exception):
        pass
    
    def __init__(self):
        self.program = []
        self.labels = {}

	# matches 0-255, used for immediate, zero page, relative modes
        byte_match = "\$([0-9A-F]{1,2})"
        # matches 0-65535, used for absolute, abs,x, abs,y, jmp indirect modes
        word_match = "\$([0-9A-F]{3,4})"

        # Instruction -> (Opcode, # of operands)
        self.opcodes_dict = ({
            "^BRK$": (0,1),                     	# BRK
            "^ORA \(%s,X\)$" % byte_match: (1,1),   	# ORA ($LL,X)
            "^ORA %s$" % byte_match: (5,1),         	# ORA $LL
            "^ASL %s$" % byte_match: (6,1),         	# ASL $LL
            "^PHP$": (8,0),                     	# PHP
	    "^ORA #%s$" % byte_match: (9,0),		# ORA #
	    "^ASL$": (10,0),				# ASL A
	    "^ORA %s$" % word_match: (13,2),		# ORA $LLHH
	    "^ASL %s$" % word_match: (14,2),		# ASL $LLHH
	    "^BPL %s$" % byte_match: (16,1),		# BPL $BB
	    "^ORA \(%s\),Y$" % byte_match: (17,1), 	# ORA ($LL),Y
	    "^ORA %s,X$" % byte_match: (21,1),		# ORA $LL,X
	    "^ASL %s,X$" % byte_match: (22,0), 		# ASL $LL,X
	    "^CLC$": (24,0),				# CLC
	    "^ORA %s,Y$" % word_match: (25,2),		# ORA $LLHH,Y
	    "^ORA %s,X$" % word_match: (29,2),		# ORA $LLHH,X
	    "^ASL %s,X$" % word_match: (30,2),		# ASL $LLHH,X
	    "^JSR %s$" % word_match: (32,2),		# JSR $LLHH
	    "^AND \(%s,X\)$" % byte_match: (33,1),	# AND ($LL,X)
	    "^BIT %s$" % byte_match: (36,1),		# BIT $LL
	    "^AND %s$" % byte_match: (37,1),		# AND $LL
	    "^ROL %s$" % byte_match: (38,1),		# ROL $LL
	    "^PLP$": (40,0),				# PLP
	    "^AND #%s$" % byte_match: (41,1),		# AND #
	    "^ROL$": (42,0),				# ROL A
	    "^BIT %s$" % word_match: (44,2),		# BIT $LLHH
	    "^AND %s$" % word_match: (45,2),		# AND $LLHH
	    "^ROL %s$" % word_match: (46,2),		# ROL $LLHH
	    "^BMI %s$" % word_match: (48,1),		# BMI $BB
	    "^AND \(%s\),Y$" % byte_match: (49,1),	# AND ($LL),Y
	    "^AND %s,X$" % byte_match: (53,1),		# AND $LL
	    "^ROL %s,X$" % byte_match: (54,1),		# ROL $LL,X
	    "^SEC$": (56,0),				# SEC
	    "^AND %s,Y$" % word_match: (57,2),		# AND $LLHH,Y
	    "^AND %s,X$" % word_match: (61,2),		# AND $LLHH,X
	    "^ROL %s,X$" % word_match: (62,2),		# ROL $LLHH,X
	    "^RTI$": (64,0),				# RTI
	    "^EOR \(%s,X\)$" % byte_match: (65,1),	# EOR ($LL,X)
	    "^EOR %s$" % byte_match: (69,1),		# EOR $LL
	    "^LSR %s$" % byte_match: (70,1),		# LST $LL
	    "^PHA$": (72,0),				# PHA
	    "^EOR #%s$" % byte_match: (73,1),		# EOR #
	    "^LSR$": (74,0),				# LSR A
	    "^JMP %s$" % word_match: (76,2),		# JMP $LLHH
	    "^EOR %s$" % word_match: (77,2),		# EOR $LLHH
	    "^LSR %s$" % word_match: (78,2),		# LSR $LLHH
	    "^BVC %s$" % byte_match: (80,1),		# BVC $BB
	    "^EOR \(%s\),Y$" % byte_match: (81,1),	# EOR ($LL),Y
	    "^EOR %s,X$" % byte_match: (85,1),		# EOR $LL,X
	    "^LSR %s,X$" % byte_match: (86,1),		# LSR $LL,X
	    "^CLI$": (88,0),				# CLI
	    "^EOR %s,Y$" % word_match: (89,2),		# EOR $LLHH,Y
	    "^EOR %s,X$" % word_match: (93,2),		# EOR $LLHH,X
	    "^LSR %s,X$" % word_match: (94,2),		# LSR $LLHH,X
	    "^RTS$": (96,0),				# RTS
	    "^ADC \(%s,X\)$" % byte_match: (97,1),	# ADC ($LL,X)
	    "^ADC %s$" % byte_match: (101,1),		# ADC $LL
	    "^ROR %s$" % byte_match: (102,1),		# ROR $LL
	    "^PLA$": (104,0),				# PLA
	    "^ADC #%s$" % byte_match: (105,1),		# ADC #
	    "^ROR$": (106,0),				# ROR A
	    "^JMP \(%s\)$": (108,2),			# JMP ($LLHH)
	    "^ADC %s$" % word_match: (109,2),		# ADC $LLHH
	    "^ROR %s$" % word_match: (110,2),		# ROR $LLHH
	    "^BVS %s$" % byte_match: (112,1),		# BVS $BB
	    "^ADC \(%s\),Y$" % byte_match: (113,1),	# ADC ($LL),Y
	    "^ADC %s,X$" % byte_match: (117,1),		# ADC $LL,X
	    "^ROR %s,X$" % byte_match: (118,1),		# ROR $LL,X
	    "^SEI$": (120,0),				# SEI
	    "^ADC %s,Y$" % word_match: (121,2),		# ADC $LLHH,Y
	    "^ADC %s,X$" % word_match: (125,2),		# ADC $LLHH,X
	    "^ROR %s,X$" % word_match: (126,2),		# ROR $LLHH,X
	    "^STA \(%s,X\)$" % byte_match: (129,1),	# STA ($LL,X)
	    "^STY %s$" % byte_match: (132,1),		# STY $LL
	    "^STA %s$" % byte_match: (133,1),		# STA $LL
	    "^STX %s$" % byte_match: (134,1),		# STX $LL
	    "^DEY$": (136,0),				# DEY
	    "^TXA$": (138,0),				# TXA
	    "^STY %s$" % word_match: (140,2),		# STY $LLHH
	    "^STA %s$" % word_match: (141,2),		# STA $LLHH
	    "^STX %s$" % word_match: (142,2),		# STX $LLHH
	    "^BCC %s$" % byte_match: (144,1),		# BCC $BB
	    "^STA \(%s\),Y$" % byte_match: (145,1),	# STA ($LL),Y
	    "^STY %s,X$" % byte_match: (148,1),		# STY $LL,X
	    "^STA %s,X$" % byte_match: (149,1),		# STA $LL,X
	    "^STX %s,Y$" % byte_match: (150,1),		# STX $LL,Y
	    "^TYA$": (152,0),				# TYA
	    "^STA %s,Y$" % word_match: (153,2),		# STA $LLHH,Y
	    "^TXS$": (154,0),				# TXS
	    "^STA %s,X$" % word_match: (157,2),		# STA $LLHH,X
	    "^LDY #%s$" % byte_match: (160,1),		# LDY #
	    "^LDA \(%s,X\)$" % byte_match: (161,1),	# LDA ($LL,X)
	    "^LDX #%s$" % byte_match: (162,1),		# LDX #
	    "^LDY %s$" % byte_match: (164,1),		# LDY $LL
	    "^LDA %s$" % byte_match: (165,1),		# LDA $LL
	    "^LDX %s$" % byte_match: (166,1),		# LDX $LL
	    "^TAY$": (168,0),				# TAY
	    "^LDA #%s$" % byte_match: (169,1),		# LDA #
	    "^TAX$": (170,0),				# TAX
	    "^LDY %s$" % word_match: (172,2),		# LDY $LLHH
	    "^LDA %s$" % word_match: (173,2),		# LDA $LLHH
	    "^LDX %s$" % word_match: (174,2),		# LDX $LLHH
	    "^BCS %s$" % byte_match: (176,1),		# BCS $BB
	    "^LDA \(%s\),Y$" % byte_match: (177,1),	# LDA ($LL),Y
	    "^LDY %s,X$" % byte_match: (180,1),		# LDY $LL,X
	    "^LDA %s,X$" % byte_match: (181,1),		# LDA $LL,X
	    "^LDY %s,Y$" % byte_match: (182,1),		# LDY $LL,Y
	    "^CLV$": (184,0),				# CLV
	    "^LDA %s,Y$" % word_match: (185,2),		# LDA $LLHH,Y
	    "^TXS$": (186,0),				# TXS
	    "^LDY %s,X$" % word_match: (188,2),		# LDY $LLHH,X
	    "^LDA %s,X$" % word_match: (189,2),		# LDA $LLHH,X
	    "^LDX %s,Y$" % word_match: (190,2),		# LDX $LLHH,Y
	    "^CPY #%s$" % byte_match: (192,1),		# CPY #
	    "^CMP \(%s,X\)$" % byte_match: (193,1),	# CMP ($LL,X)
	    "^CPY %s$" % byte_match: (196,1),		# CPY $LL
	    "^CMP %s$" % byte_match: (197,1),		# CMP $LL
	    "^DEC %s$" % byte_match: (198,1),		# DEC $LL
	    "^INY$": (200,0),				# INY
	    "^CMP #%s$" % byte_match: (201,1),		# CMP #
	    "^DEX$": (202,0),				# DEX
	    "^CPY %s$" % word_match: (204,2),		# CPY $LLHH
	    "^CMP %s$" % word_match: (205,2),		# CMP $LLHH
	    "^DEC %s$" % word_match: (206,2),		# DEC $LLHH
	    "^BNE %s$" % byte_match: (208,1),		# BNE $BB
	    "^CMP \(%s\),Y$" % byte_match: (209,2),	# CMP ($LL),Y
	    "^CMP %s,X$" % byte_match: (213,1),		# CMP $LL,X
	    "^DEC %s,X$" % byte_match: (214,1),		# DEC $LL,X
	    "^CLD$": (216,1),				# CLD
	    "^CMP %s,Y$" % word_match: (217,2),		# CMP $LLHH,Y
	    "^CMP %s,X$" % word_match: (221,2),		# CMP $LLHH,X
	    "^DEC %s,X$" % word_match: (222,2),		# DEC $LLHH,X
	    "^CPX #%s$" % byte_match: (224,1),		# CPX #
	    "^SBC \(%s,X\)$" % byte_match: (225,1),	# SBC ($LL,X)
	    "^CPX %s$" % byte_match: (228,1),		# CPX $LL
	    "^SBC %s$" % byte_match: (229,1),		# SBC $LL
	    "^INC %s$" % byte_match: (230,1),		# INC $LL
	    "^INX$": (232,0),				# INX
	    "^SBC #%s$" % byte_match: (233, 1),		# SBC #
	    "^NOP$": (234,0),				# NOP
	    "^CPX %s$" % word_match: (236,2),		# CPX $LLHH
	    "^SBC %s$" % word_match: (237,2),		# SBC $LLHH
	    "^INC %s$" % word_match: (238,2),		# INC $LLHH
	    "^BEQ %s$" % byte_match: (240,1),		# BEQ $BB
	    "^SBC \(%s\),Y$" % byte_match: (241,1),	# SBC ($LL),Y
	    "^SBC %s,X$" % byte_match: (245,1),		# SBC $LL,X
	    "^INC %s,X$" % byte_match: (246,1),		# INC $LL,X
	    "^SED$": (248,0),				# SED
	    "^SBC %s,Y$" % word_match: (249,2),		# SBC $LLHH,Y
	    "^SBC %s,X$" % word_match: (253,2),		# SBC $LLHH,X
	    "^INC %s,X$" % word_match: (254,2)		# INC $LLHH,X
            })

    # Load program from a file, converts it to uppercase, remove blank lines
    # and split it into a list of code lines
    def load_program(self, path):
        f = open(path)
        buff = f.read() # read from program file
        buff = buff.upper() # convert to uppercase
        buff = buff.replace("\n\n", "\n") # remove empty lines
        self.program = buff.split("\n") # split each instruction into a list

        for i in range(len(self.program)):
            self.program[i] = self.program[i].strip()
       
        f.close()
        

    # Convert n into an hexadecimal string
    def hexadecimal(self, n, word=False):
        symbols = "0123456789ABCDEF"
        result = ""

        if n < 0:
            n %= 256
        elif n == 0:
            return "$00"
        
        while n > 0:
            result = symbols[n % 16] + result
            n //= 16

        return ("$00" + result if word and len(result) < 3 else
            "$0" + result if len(result) == 1 or len(result) == 3 else "$" + result)

    def is_label(self, s):
        return s[-1] == ":"

    def is_reserved(semf, s):
        return ".ORG" in s or ".DB" in s or ".DW" in s

    # Get the length of an instruction, if the instruction is unknow, an error is thrown
    def get_instruction_length(self, instruction):
        for op in self.opcodes_dict:
            match = re.match(op, instruction)
            if match:
                return self.opcodes_dict[op][1] + 1

        return 0

    def process(self):

        address = 0
        delete_list = [] # Stores the instructions of the program list to delete
        
        # pass 1: replace bin and dec with hex values,
        # and get addresses from labels + delete labels 

        for i in range(len(self.program)):
            
            instruction = self.program[i]
            
            search = re.search(" %([0-1]+)$", instruction)
            if search:
                binary = search.group(1)
                decimal = int(search.group(1),2)
                instruction = self.program[i].replace("%" + binary, self.hexadecimal(decimal))
            
            search = re.search(" #?(\d+)$", instruction)
            if search:
                decimal = int(search.group(1))
                instruction = self.program[i].replace(str(decimal), self.hexadecimal(decimal))

            # Set address to x if instruction = .org x
            search = re.search("^.ORG \$([0-9A-F]{1,4})$", instruction)
            if search:
                org = int(search.group(1),16)
                address = org
                delete_list.append(instruction)

            # Increment address counter
            address += self.get_instruction_length(instruction)

            # If the instruction is a label identifier, save its address and delete the instruction from the program
            if self.is_label(instruction):
                # Raise exception if label already present in the label dict
                if instruction[:-1] in self.labels:
                    raise self.AssemblyException("\"" + instruction[:-1] + "\" identifier used multiple times.")
                self.labels[instruction[:-1]] = address
                delete_list.append(instruction)

            # Rewrite the instruction with the right hexadecimal operands
            self.program[i] = instruction

        # Delete the instructions of the program stored in the delete list
        for d in delete_list:
            self.program.remove(d)

        

        # pass 2: replace labels in branching instructions with relative
        # offset and labels in other instructions with absolute address

        address = 0

        for i in range(len(self.program)):
            instruction = self.program[i]
            for lab in self.labels:
                if lab in instruction:
                    
                    # Fetch instruction
                    search = re.search("^(\w{3}) \w+", instruction)
                    if search:
                        inst = search.group(1)
                        # Branching conditions
                        if inst in ["BEQ", "BNE", "BVC", "BVS", "BCC", "BCS", "BPL", "BMI"]:
                            offset = self.labels[lab] - address - 2
                            if offset < -128 or offset > 127:
                                raise self.AssemblyException("\"" + instruction + "\": \"" + lab + "\" out of range") 
                            instruction = instruction.replace(lab, self.hexadecimal(offset))
                        else:
                            instruction = instruction.replace(lab, self.hexadecimal(self.labels[lab],word=True))
                self.program[i] = instruction
                # Increment address counter    
                address += self.get_instruction_length(instruction)
                
        

    # Takes an instruction string as parameter
    # Returns the corresponding opcode, and the eventual low and high byte operands in a list
    def assemble_instruction(self, inst):
        result = []
        for op in self.opcodes_dict:
            search = re.search(op, inst)
            if search:
                value = self.opcodes_dict[op]
                result.append(value[0]) # append instruction's opcode
                if value[1] == 1 and not inst == "BRK":
                    operand = int(search.group(1),16)
                    result.append(operand)
                elif value[1] == 2:
                    operand = int(search.group(1),16)
                    result.append(operand & 255)
                    result.append(operand >> 8)
                break
            
        return result

    # Converts the processed program into a list of opcodes
    # Prints the encoded program if debug is enabled
    def assemble_program(self, path="output.bin", debug=False):
        result = []
        self.process() # converts binary and decimal constants into hexacimal and parse labels from the program
        # iterate through the processed program
        for inst in self.program:
            assembled_instruction = self.assemble_instruction(inst)
            if not assembled_instruction:
                raise self.AssemblyException("Syntax Error: " + inst)
            if debug:
                print(" ".join(list(map(self.hexadecimal, assembled_instruction))))
            result += assembled_instruction

        f = open(path, "w+b")
        f.write(bytearray(result))
        f.close()
        print("Done.")
        
        self.listing()
        

    # Prints the label dictionnary once the program has been encoded
    def listing(self):
        print("\nLabels list :")
        for lab in self.labels:
            print(lab + " - " + self.hexadecimal(self.labels[lab]))

a = Assembler()
a.load_program("code.asm")
a.assemble_program(debug=True)
